html {
    color: #000; 
    background: #fefefe;
    font-family: serif;
    margin: 0;
    border: 0;
    padding: 0;
    line-height: 120%;
}

body {
    color: #000; 
    background: #fefefe;
    margin: 10px;
    border: 0;
    padding: 0;
}

h1, h2, h3 {
    color: #000; 
    background: #fefefe;
    text-align: left; 
    font-family: sans-serif;
}

h1 {
    font-size: 150%;
    line-height: 150%;
}

h2 {
    font-size: 130%;
}

h3 {
    font-size: 100%;
}

a:link {
    color: #1b56ce;
    font-weight: bold;
}

a:visited {
    color:#1b56ce;
}

a:link:active, a:link:visited {
    color:#ff0000;
}

a:link:hover, a:visited:hover {
    color: #d81e1e;
}

.link {
    color: #1b56ce;
    font-weight: bold;
    text-decoration: underline;
}

.link abbr {
    color: #1b56ce;
    font-weight: bold;
    text-decoration: underline;
}

.link abbr:hover {
    color: #1b56ce;
    font-weight: bold;
    text-decoration: underline;
}


.link span:hover {
    color: #d81e1e;
}


a.submitter:link {
    color: #242424;
    font-family: sans-serif;
    font-size: 95%;
    text-decoration: underline;
    font-weight: normal;
}

a.submitter:visited, a.submitter:active {
    color: #6e6e6e;
    font-family: sans-serif;
    font-size: 95%;
}

a.submitter:hover, a.submitter:visited:hover {
    color: #d01414;
    font-family: sans-serif;
    font-size: 95%;
}

pre.message {
    font-family: monospace;
    padding-top: 8px;
    margin-top: 0;
    border-top: 0;
    white-space: pre-wrap;
}

pre.wrapping {
    width: 80ch;
}

.sparse li {
    padding-top: 1ex;
    margin-top: 1ex;
    border-top: 1ex;
}

a.bugtitle {
    font-weight: bold;
    font-size: 110%;
}


pre.headers {
    font-family: sans-serif;
    font-size: 95%;
    color: #3c3c3c;
    background-color: #f0f0f0;
    padding: 2px;
    border: #a7a7a7 1px solid;
    line-height: 120%
}

div.headers { 
    border: #a7a7a7 1px solid;
    background-color: #f0f0f0;
    padding: 2px;
    overflow: auto;
 }

div.header {
    font-family: sans-serif;
    font-size: 95%;
    color: #3c3c3c;
    padding: 0px;
    line-height: 120%;
    margin: 0px;
}

div.headers p {
    font-family: sans-serif;
    font-size: 95%;
    color: #3c3c3c;
    padding: 0px;
    line-height: 120%;
    margin: 0px;
}

div.headers img {  
    float: right;
}

span.headerfield { 
  font-weight: bold
 };

pre.mime {
    font-family: monospace;
    font-size: 95%;
    color: #797979;
}

/* This must be separate from the other CSS to make the showing of
   unimportant messages work in bugreport.cgi. */ 
.infmessage { display: none; }

.infmessage {
    font-family: sans-serif;
    font-size: 90%;
    color: #686868;
}

.msgreceived {
    font-family: sans-serif;
    font-size: 90%;
    color: #686868;
}

.msgreceived p {
    width: 120ch;
    margin-top: 0px;
    margin-bottom: 0px;
}

p.msgreceived {
    width: 120ch;
    margin-top: 0px;
    margin-bottom: 0px;
}

.buginfo p
{ 
  font-family: sans-serif;
  font-size: 110%;
  margin-bottom: 0px
}

.buginfo p + p
{ 
  margin: 0;
  margin-top: 0px;
  border: 0;
}


.pkginfo p
{ font-family: sans-serif;
  font-size: 110%;
  margin-bottom: 0px
}

.pkginfo p + p
{ 
  margin: 0;
  margin-top: 0px;
  padding: 0;
  border: 0;
}


.versiongraph
{
  float: right
  
}

pre.tags {
    color: #a3a3a3;
    font-size: 90%;
}

hr.thin {
    color: #a1a1a1;
}

em.severity {
    color: #c31212;
}

code, address {
    font-family: sans-serif;
}

p {
    font-family: sans-serif;
    font-size: 95%;
}

h2.outstanding {
    font-family: sans-serif;
    background-color: #f0f0f0;
    color: #3c3c3c;
    border: #a7a7a7 1px solid;
    padding: 10px;
}

a.options:link, a.options:visited {
    font-family: sans-serif;
    background-color: #f0f0f0;
    color: #3c3c3c;
    text-decoration: none;
    border-bottom: #3c3c3c 1px dotted;
}

li {
    list-style-type: square;
}

.shortbugstatus
{ 
  font-family: sans-serif;
  
  }

.shortbugstatusextra
{ font-family: sans-serif;
  margin: 5px;
  margin-top: 2px;
  padding: 5px;
  /* display: none; */
  /* z-index: 1; */
  /* position: absolute; */
  left: 120px;
  background-color: #ffffff;
/*  border: #000 1px solid; */
  position: static;
  display: block;
  border: 0;
  }

.shortbugstatusextra span
{ margin: 0;
  margin-top: 0px;
  padding: 0;
  border: 0;
  display: inline;
  }

#status_mask {
    -moz-opacity: 0;
    opacity: 0;
    filter: alpha(opacity=0);
    z-index: 1;
    position: absolute;
    width: 100%;
    height: 100%;
    top: 0px;
    left: 0px;
    display: none;
}

.bugs li {
    margin-top: 5px;
}

input {
    border: #000 1px solid;
    margin: 3px;
}

table.forms {
    font-size: 95%;
    font-family: sans-serif;
    margin-left: 10px;
}

select {
    margin: 3px;
    border: #000 1px solid;
}

